(******************************************************************************)
(*   #####   #####    ####      Formalization of the GPS program logic        *)
(*  ##       ##  ##  ##         by Turon, Vafeiadis and Dreyer                *)
(*  ##  ###  #####    ####                                                    *)
(*  ##   ##  ##          ##     Copyright (c) Viktor Vafeiadis, 2013          *)
(*   #####   ##       ####      See README.txt for license                    *)
(******************************************************************************)

Require Import Vbase Relations ClassicalDescription ClassicalChoice.
Require Import List Permutation Sorting extralib. 
Require Import c11 exps.
Require Import GpsModel GpsModelLemmas.

Local Open Scope string_scope.

Set Implicit Arguments.

(******************************************************************************)
(******************************************************************************)
(** * 2.3. Global safety *) 
(******************************************************************************)
(******************************************************************************)

(******************************************************************************)
(** ** 2.3.1. Domains *) 
(******************************************************************************)

Inductive tagged_edge :=
  | Tsb (a b : actid)
  | TsbS (a: actid)
  | Trf (a b : actid)
  | TrfS (a: actid)
  | Tesc (a b : actid)
  | TescS (a: actid)
  | TcondS (a: actid).

Definition labeling := tagged_edge -> resource. 

Definition escrow_intros := list (escrowTy * resource).

Definition tagged_threads := list (actid * exp * resource * (nat -> RProp)).

Definition te_src te :=
  match te with
    | Tsb  a _ 
    | TsbS a
    | Trf a _ 
    | TrfS a
    | Tesc a _
    | TescS a
    | TcondS a => a
  end.

Definition te_tgt te :=
  match te with
    | Tsb  _ b 
    | Trf _ b 
    | Tesc _ b => Some b
    | TsbS _
    | TrfS _
    | TescS _
    | TcondS _ => None
  end.


(******************************************************************************)
(** ** 2.3.2. Valid ghost moves *) 
(******************************************************************************)

Definition valid_ghost r r' (I : escrow_intros) :=
  exists p g escr,
    r = Rdef p g escr /\
    exists g' escr',
      r' = Rdef p g' escr' /\
      (*(exists g_rem, gres_plus (gres_strip g) g_rem = Some g') /\ *)
      (forall sigma, escr' sigma <-> escr sigma \/ exists rr, In (sigma, rr) I).  

(******************************************************************************)
(** ** 2.3.2. Valid edge labels for a node *) 
(******************************************************************************)

Notation "x +!+ y" := (res_plus x y) (at level 55, right associativity).

Definition in_sb  acts L b := rsum (map L (map (Tsb^~ b) acts)).
Definition in_rf  acts L b := rsum (map L (map (Trf^~ b) acts)).
Definition in_esc acts L b := rsum (map L (map (Tesc^~ b) acts)).
Definition out_sb  acts L b := L (TsbS b) +!+ rsum (map L (map (Tsb b) acts)).
Definition out_rf  acts L b := L (TrfS b) +!+ rsum (map L (map (Trf b) acts)).
Definition out_esc acts L b := L (TescS b) +!+ rsum (map L (map (Tesc b) acts)).

Definition incoming acts L b := 
  in_sb acts L b +!+ in_rf acts L b +!+ in_esc acts L b.

Definition outgoing acts L b := 
  out_sb acts L b +!+ out_rf acts L b +!+ out_esc acts L b +!+ L (TcondS b).

Definition escr_ffun L acts sr :=
  if excluded_middle_informative 
    (exists P P', 
       interp_escr (fst sr) = (P, P') /\
       P' (snd sr) /\
       (forall c (IN : In c acts) (SAT: P (L (TcondS c))), False))
  then true
  else false.

Definition valid_node  (acts N : list actid) 
           (amap : actid -> act)
           (rf   : actid -> option actid) 
           (L : labeling) (b: actid) :=
  exists r I,
    << VG : valid_ghost (incoming acts L b) (r +!+ out_esc acts L b +!+ L (TcondS b)) I >> /\ 
    << GUAR: guar (in_sb acts L b +!+ in_rf acts L b) r (amap b) 
                  (out_sb acts L b) (out_rf acts L b) >> /\
    << RS : res_lt (res_strip (incoming acts L b)) r >> /\
    << VI  : forall sigma r (GIN: In (sigma, r) I), 
            exists P P', interp_escr sigma = (P, P') /\ P' r >> /\
    << VrfM: forall c (INc: In c N) (RF: rf c = Some b) (RMW: is_rmw (amap c)), 
             L (Trf b c) = out_rf acts L b >> /\
    << VrfS: (forall c (INc: In c N) (RF: rf c = Some b) (RMW: is_rmw (amap c)), False) -> 
             L (TrfS b) = out_rf acts L b >> /\ 
    << VrfS': res_strip (L (TrfS b)) = res_strip (out_rf acts L b) >> /\
    << VescS: res_lt (rsum (map snd (filter (escr_ffun L acts) I))) (L (TescS b)) >>.

(******************************************************************************)
(** ** 2.3.4. Concurrent compatibility *) 
(******************************************************************************)

Definition may_happens_before amap sb rf :=
  clos_refl_trans _ (fun a b => sb a b \/ synchronizes_with amap rf a b).

Definition compat amap sb rf L :=
  forall E 
    (NDUP: NoDup E)
    (IND: forall te (IN: In te E) te' (IN': In te' E) b 
           (TGT: te_tgt te' = Some b) 
           (HB: may_happens_before amap sb rf b (te_src te)),
       False),
  rsum (map L E) <> Rundef.
 
(******************************************************************************)
(** ** 2.3.5. Protocol conformance *)
(******************************************************************************)

Definition is_atomic_writeL a loc :=
  match a with
    | Askip | Aalloc _ _  | Aload _ _ _  | Astore ATna _ _=> False 
    | Astore ATat l _ | Armw l _ _ => l = loc
  end. 

Definition conform acts amap (mo : actid -> actid -> Prop) L N :=
  forall loc 
    a (INa : In a N) (ATa : is_atomic_writeL (amap a) loc)
    b (INb: In b N)  (Wb : is_writeL (amap b) loc) (MO: mo a b),
    prot_trans_ok (res_get (out_rf acts L a) loc)
                  (res_get (out_rf acts L b) loc).
 
(******************************************************************************)
(** ** 2.3.6. Global safety *)
(******************************************************************************)

Definition label_restr acts amap (sb : relation actid) rf L :=
  << LRsb  : forall a b, ~ sb a b -> L (Tsb a b) = res_emp >> /\
  << LRrf  : forall a b, ~ synchronizes_with amap rf a b -> L (Trf a b) = res_emp >> /\
  << LResc : forall a b, ~ happens_before amap sb rf a b -> L (Tesc a b) = res_emp >> /\
  << LRsbS : forall a, ~ In a acts -> L (TsbS a) = res_emp >> /\
  << LRrfS : forall a, ~ In a acts -> L (TrfS a) = res_emp >> /\
  << LRescS : forall a, ~ In a acts -> L (TescS a) = res_emp >> /\
  << LRcondS: forall a, ~ In a acts -> L (TcondS a) = res_emp >>.

Definition gsafe n acts amap sb rf mo L (T : tagged_threads) :=
  << V : forall a (IN: In a acts), valid_node acts acts amap rf L a >> /\
  << CC: compat amap sb rf L >> /\
  << CO: conform acts amap mo L acts >> /\
  << SBS : forall a (IN: In a acts), 
             L (TsbS a) = 
             rsum (map (fun aerP => snd (fst aerP))
                       (filter (fun aerP => fst (fst (fst aerP)) == a) T)) >> /\
  << CT : forall aerP (IN: In aerP T), In (fst (fst (fst aerP))) acts >> /\
  << SAFE : forall a e r PP (IN: In (a, e, r, PP) T), safe n e PP r >> /\
  << CONS: Consistent acts amap sb rf mo >> /\
  << ND  : NoDup acts >> /\
  << LR  : label_restr acts amap sb rf L >>.


(******************************************************************************)
(** ** Helper lemmas *)
(******************************************************************************)

(** Prefix closure *)

Definition prefix_closed (sb : actid -> actid -> Prop) rf (N : list actid) :=
  << PCsb : forall b (IN: In b N) a (SB: sb a b), In a N >> /\
  << PCrf : forall b (IN: In b N) a (RF: rf b = Some a), In a N >>.

Definition pres_in (sb : actid -> actid -> Prop) rf b N :=
  (forall a (SB: sb a b), In a N) /\
  (forall a (SB: rf b = Some a), In a N).

Lemma prefix_closed_hb amap sb rf l a b :
  prefix_closed sb rf l ->
  happens_before amap sb rf a b ->
  In b l ->
  In a l.
Proof.  
  induction 2; eauto.
  unfold imm_happens_before, synchronizes_with, prefix_closed in *; desf; eauto.
Qed.

Lemma pres_in_hb amap sb rf l a b :
  prefix_closed sb rf l ->
  happens_before amap sb rf a b ->
  pres_in sb rf b l ->
  In a l.
Proof.  
  ins; apply clos_trans_tn1 in H0; destruct H0;
    [|eapply prefix_closed_hb; eauto; try eby eapply clos_tn1_trans];
    unfold imm_happens_before, synchronizes_with, pres_in in *; desf; eauto.
Qed.

Lemma Consistent_pres_in :
  forall b acts amap sb rf mo (CONS: Consistent (b::acts) amap sb rf mo),
    pres_in sb rf b acts.
Proof.
  red; unfold NW; ins; red in CONS; desf; split; ins.
    destruct (proj1 (proj2 FIN _ _ SB)); ins; desf; edestruct ACYC; vauto.
  specialize (Crf b); rewrite SB in *; desf.
  assert (X: In a (b::acts)) by (by apply (proj1 FIN); intro X; rewrite X in *).
  destruct X; desf; edestruct ACYC; vauto.
Qed.

Lemma Consistent_prefix_closed :
  forall acts amap sb rf mo (CONS: Consistent acts amap sb rf mo),
    prefix_closed sb rf acts.
Proof.
  red; unfold NW; ins; red in CONS; desf; split; ins.
    by apply (proj2 FIN) in SB; desf.
  specialize (Crf b); rewrite RF in *; desf.
  by apply (proj1 FIN); intro X; rewrite X in *.
Qed.

Lemma Consistent_prefix_closed2 :
  forall b acts amap sb rf mo (CONS: Consistent (b :: acts) amap sb rf mo)
    (Nsb: forall x, sb b x -> False) 
    (Nrf: forall x, rf x = Some b -> False),
    prefix_closed sb rf acts.
Proof.
  ins; apply Consistent_prefix_closed in CONS.
  red; unfold NW; ins; red in CONS; desf; split; ins.
    by edestruct PCsb; eauto; desf; exfalso; eauto.
  by edestruct PCrf; eauto; desf; exfalso; eauto.
Qed.


(** *** Lemmas about label restrictions *)
(******************************************************************************)

Lemma lr_upd_sbS acts amap sb rf L b r :
  label_restr acts amap sb rf L -> In b acts ->
  label_restr acts amap sb rf (upd L (TsbS b) r).
Proof.
  destruct 1; desf; repeat split; unfold NW, upd; ins; desf; auto.
Qed.

Lemma lr_upd_rfS acts amap sb rf L b r :
  label_restr acts amap sb rf L -> In b acts ->
  label_restr acts amap sb rf (upd L (TrfS b) r).
Proof.
  destruct 1; desf; repeat split; unfold NW, upd; ins; desf; auto.
Qed.

Lemma lr_upd_escS acts amap sb rf L b r :
  label_restr acts amap sb rf L -> In b acts ->
  label_restr acts amap sb rf (upd L (TescS b) r).
Proof.
  destruct 1; desf; repeat split; unfold NW, upd; ins; desf; auto.
Qed.

Lemma lr_upd_condS acts amap sb rf L b r :
  label_restr acts amap sb rf L -> In b acts ->
  label_restr acts amap sb rf (upd L (TcondS b) r).
Proof.
  destruct 1; desf; repeat split; unfold NW, upd; ins; desf; auto.
Qed.

Lemma lr_upd_sb acts amap sb rf L a b r :
  label_restr acts amap sb rf L -> sb a b ->
  label_restr acts amap sb rf (upd L (Tsb a b) r).
Proof.
  destruct 1; desf; repeat split; unfold NW, upd; ins; desf; auto.
Qed.

Lemma lr_upd_rf acts amap sb rf L a b r :
  label_restr acts amap sb rf L -> synchronizes_with amap rf a b ->
  label_restr acts amap sb rf (upd L (Trf a b) r).
Proof.
  destruct 1; desf; repeat split; unfold NW, upd; ins; desf; auto.
Qed.

Lemma lr_upd_esc acts amap sb rf L a b r :
  label_restr acts amap sb rf L -> happens_before amap sb rf a b ->
  label_restr acts amap sb rf (upd L (Tesc a b) r).
Proof.
  destruct 1; desf; repeat split; unfold NW, upd; ins; desf; auto.
Qed.

Hint Resolve lr_upd_sbS lr_upd_rfS lr_upd_escS lr_upd_condS
  lr_upd_sb lr_upd_rf lr_upd_esc.



(** *** Various lemmas *)
(******************************************************************************)

Lemma Permutation_rsum l l' : Permutation l l' -> rsum l = rsum l'.
Proof.
  induction 1; ins; try congruence; apply res_plusAC.
Qed.

Lemma Permutation_ffun l l' (P: Permutation l l') L :
  escr_ffun L l = escr_ffun L l'.
Proof.
  extensionality x; unfold escr_ffun; desf; destruct n; desf;
  eauto 10 using Permutation_in, Permutation_sym.
Qed.

Lemma Permutation_valid_node l l' (P: Permutation l l') amap M rf L a :
  valid_node l' amap M rf L a ->
  valid_node l amap M rf L a.
Proof.
  unfold valid_node; ins; desf; unfold NW; exists r, I.
  unfold incoming, in_sb, in_rf, in_esc, out_sb, out_rf, out_esc.
  rewrite !(Permutation_rsum (Permutation_map _ (Permutation_map _ P))). 
  rewrite !(Permutation_ffun P). 
  repeat (split; try done).
Qed.

Lemma escr_ffun_upd_irr L te r :
  (forall a, te <> TcondS a) ->
  escr_ffun (upd L te r) = escr_ffun L. 
Proof.
  ins; extensionality acts; extensionality x; 
  unfold escr_ffun; desf; destruct n; desf; repeat eexists; eauto;
  ins; eapply (e1 c); unfold upd in *; desf; edestruct H; ins.
Qed.

Lemma helper1 a b c : 
  res_strip a +!+ b = Rundef ->
  res_lt a c ->
  res_lt b c ->
  c = Rundef.
Proof.
  ins; rewrite <- (res_plus_strip_idemp c).
  unfold res_lt in *; ins; desf.
  case_eq (a +!+ r_rem0) as X; try done.
  rewrite <- H1 at 1; rewrite <- X, res_strip_plus; try congruence. 
  by rewrite res_plusA, !(res_plusAC (res_strip a)), <- res_plusA, H.
Qed.

Lemma replace_out_rf :
  forall a b c L acts
    (NEQ: c <> b)
    (ND: NoDup acts)
    (EMP: L (Trf a b) = res_emp),
  out_rf acts (upd L (Trf a b) (res_strip (out_rf acts L a))) c = out_rf acts L c.
Proof.
  ins; unfold out_rf.
  destruct (classic (In b acts)); desf.
    destruct (classic (a = c)); desf.
      eapply In_NoDup_Permutation in H; eauto; desf.
      rewrite !(Permutation_rsum (Permutation_map _ (Permutation_map _ H))); ins.
      rewrite ?upds, !map_upd_irr, ?updo; ins; try by rewrite in_map_iff; intro; desf.
      by rewrite EMP, res_plus_emp_l, res_plusAC, res_plusC, res_plus_strip_idemp.
    by rewrite map_upd_irr, updo; ins; rewrite in_map_iff; intro; desf.
  by rewrite map_upd_irr, updo; ins; rewrite in_map_iff; intro; desf.
Qed.  

Lemma replace_out_rf2 :
  forall a b acts c L
    (ND: NoDup acts)
    (INa: In a acts)
    (INb: In b acts)
    (EQ: L (TrfS a) = out_rf acts L a),
    out_rf (acts)
           (upd (upd L (TrfS a) (res_strip (out_rf acts L a)))
           (Trf a b) (out_rf acts L a)) c = 
    out_rf acts L c.
Proof.
  ins; unfold out_rf in *.
  destruct (classic (a = c)); desf.
  2: by rewrite !map_upd_irr; rupd; ins; try rewrite in_map_iff; intro; desf.
  eapply In_NoDup_Permutation in INb; eauto; desf.
  revert EQ.
  rewrite !(Permutation_rsum (Permutation_map _ (Permutation_map _ INb))); ins.
  rewrite !map_upd_irr; rupd; try by rewrite in_map_iff; intro; desf.
  rewrite <- res_plusA, res_plus_strip_l.
  case (classic (L (TrfS c) = Rundef)); [intros -> | intro D]; ins. 
  exploit res_plus_ubound; eauto; intro X; rewrite rknow_plusE in X; desf. 
    by rewrite X0 at 2; rewrite !res_plusA, res_plus_strip_idemp.
  intro M; rewrite M in *; destruct (L (TrfS c)); ins.
Qed.

Lemma replace_edge_emp :
  forall A (a : A) f pi l 
    (EMP: f a = res_emp) 
    (ND: NoDup l)
    (IN : In a l),
  rsum (map (upd f a pi) l) = pi +!+ rsum (map f l).
Proof.
  ins; induction ND; ins. 
  desf; rewrite ?upds, ?updo, ?EMP, ?res_plus_emp_l; try congruence.
    by rewrite map_upd_irr.
  by rewrite IHND, res_plusAC.
Qed. 

Lemma is_write_weaken typ l v : is_writeLV typ l v -> is_writeL typ l.
Proof. destruct typ; ins; desf. Qed.

Lemma is_access_weaken typ l v : is_writeLV typ l v -> is_access typ.
Proof. destruct typ; ins; desf. Qed.

Hint Resolve is_write_weaken is_access_weaken.


Lemma helper2 A f (x : A) l :
  NoDup l ->
  res_strip (f x) +!+ rsum (map f l) = Rundef ->
  rsum (map f (undup (x :: l))) = Rundef.
Proof.
  ins; desf; rewrite undup_nodup; ins.
    eapply In_NoDup_Permutation in i; desf.
    rewrite (Permutation_rsum (Permutation_map f i)) in *; ins.
    by rewrite res_plusAC, <- res_plusA, res_plus_strip_idemp in H0.
  by rewrite <- (res_plus_strip_idemp (f x)), res_plusA, H0, res_plusC. 
Qed.

Definition mydecf A (f : A -> Prop) x :=
  if excluded_middle_informative (f x) then true else false.

Lemma rsum_map_filter A (f : A -> resource) h l :
  (forall x, In x l -> ~ h x -> f x = res_emp) ->
  rsum (map f (filter (mydecf h) l)) =
  rsum (map f l).
Proof.
  induction l; ins; desf; ins; f_equal; eauto.  
  rewrite H, res_plus_emp_l; eauto.
  unfold mydecf in *; desf.
Qed.

Lemma rsum_map_map_filter A B f (g : A -> B) h l :
  (forall x, In x l -> ~ h x -> f (g x) = res_emp) ->
  rsum (map f (map g (filter (mydecf h) l))) =
  rsum (map f (map g l)).
Proof.
  induction l; ins; desf; ins; f_equal; eauto.  
  rewrite H, res_plus_emp_l; eauto.
  unfold mydecf in *; desf.
Qed.

Lemma mhb_trans amap sb rf a b c :
  may_happens_before amap sb rf a b ->
  may_happens_before amap sb rf b c ->
  may_happens_before amap sb rf a c.
Proof.
  unfold may_happens_before; eauto using rt_trans.
Qed.

Lemma hb_in_ct amap sb rf a b :
  happens_before amap sb rf a b ->
  clos_trans _ (fun a b => sb a b \/ rf b = Some a) a b.
Proof.
  induction 1 as [? ? [|[]]|]; eauto using clos_trans.
Qed.

Lemma hb_sb amap (sb: _ -> _ -> Prop) rf a b :
  sb a b -> happens_before amap sb rf a b.
Proof.
  vauto. 
Qed.

Lemma hb_sw amap sb rf a b:
  synchronizes_with amap rf a b ->
  happens_before amap sb rf a b.
Proof.
  vauto.
Qed.

Lemma hb_trans amap sb rf a b c :
  happens_before amap sb rf a b ->
  happens_before amap sb rf b c ->
  happens_before amap sb rf a c.
Proof.
  unfold happens_before; eauto using t_trans.
Qed.

Hint Immediate hb_in_ct hb_sb hb_sw.

Lemma mhbE amap sb rf a b : 
  may_happens_before amap sb rf a b <->
  a = b \/ happens_before amap sb rf a b.
Proof.
  unfold may_happens_before, happens_before, imm_happens_before; split; ins; desf; eauto;
  try induction H; desf; eauto using clos_trans, clos_refl_trans.
Qed. 

Lemma no_rf_rmw_overtaken acts amap sb rf mo a b b_rf :
  Consistent acts amap sb rf mo ->
  rf b = Some b_rf ->
  rf a = Some b_rf ->
  is_rmw (amap a) ->
  happens_before amap sb rf a b ->
  False.
Proof.
  unfold Consistent; ins; desc.        
  edestruct Crmw as [MO _]; eauto. 
  generalize (Crf b), (Crf a); rewrite H0, H1; ins; desc.
  eapply Cwr in H0; eauto.
  destruct (amap b_rf); ins; destruct (amap a); ins; destruct (amap b); ins; desf. 
Qed.

Lemma no_rf_rmw_double acts amap sb rf mo a b b_rf :
  Consistent acts amap sb rf mo ->
  rf b = Some b_rf ->
  rf a = Some b_rf ->
  is_rmw (amap a) ->
  is_rmw (amap b) ->
  a = b.
Proof.
  unfold Consistent; ins; desc.
  edestruct Crmw with (a := a) as [MO _]; eauto. 
  edestruct Crmw with (a := b) as [MO' _]; eauto. 
  generalize (Crf b), (Crf a); rewrite H0, H1; ins; desc.
  assert (l0 = l) by (destruct (amap b_rf); ins; desf); subst.
  apply NNPP; intro N; eapply Cmo with (l:=l) in N; desf; try eapply Crmw in N; eauto.
    by destruct (amap a); ins; desf.
    by destruct (amap b); ins; desf.
Qed.

Lemma aguar_interp_prot r l v r_sb r_rf tau S :
  res_get r_rf l = p_at tau S ->
  atomic_guar r l v r_sb r_rf ->
  interp_prot tau (get_state S) v r_rf. 
Proof.
  unfold atomic_guar; ins; desf; rewrite H in *; desf;
  replace (get_state S') with s; try done.
    by unfold get_state; destruct S'; ins; subst; ins.

  assert (IN := proj2 (SEQ s) (in_eq _ _)).
  assert (A: forall x, In x (SL_list S') -> prot_trans tau0 x s).
    ins; apply SEQ in H0; desf; eauto using prot_trans_refl. 
  clear - IN A.
  unfold get_state; des_eqrefl; [exfalso|]; ins.
    by destruct S'; ins; destruct (exists_last SL_nemp) as (? & ? & ?); subst; 
       rewrite rev_app_distr in *.
  destruct S'; ins. 
  assert (exists l, SL_list = l ++ s0 :: nil).
    revert x EQ; clear; induction SL_list using rev_ind; ins.
    by rewrite rev_app_distr in *; ins; desf; vauto.
  desf.
  rewrite in_app_iff in *; ins; desf.
   exploit Sorted_app_end_inv; eauto; ins.
  eapply prot_trans_ord; eauto using in_or_app, in_eq.
Qed.

Lemma slist_lt_trans A (x y z: slist A) :
  slist_lt x y ->
  slist_lt y z ->
  slist_lt x z.
Proof.
  unfold slist_lt; ins; desf.
  edestruct H as (? & ? & ?); eauto.
  edestruct H0 as (? & ? & ?); eauto using prot_trans_trans.
Qed.

Lemma exists_immediate : 
  forall A (r : relation A) 
    (IRR: irreflexive r) (T: transitive _ r) a b 
    (HAS : r a b) l
    (FIN : forall c, r a c -> r c b -> In c l),
  exists a', immediate r a' b /\ (a = a' \/ r a a').
Proof.
  induction[a HAS] l; ins.
    exists a; repeat split; eauto.
  destruct (classic (r a b /\ r a0 a)); desf; eauto. 
    destruct (IHl a) as [a' ?]; ins; desf; eauto.
    by destruct (FIN c); desf; eauto; exfalso; eauto. 
  eapply (IHl a0); eauto; ins. 
  by destruct (FIN c); desf; destruct H.
Qed. 

Lemma compat_mon_upd amap sb rf L a r :
  compat amap sb rf (upd L a (L a +!+ r)) ->
  compat amap sb rf L.
Proof.
  repeat red; ins; apply (H E); auto.
  destruct (classic (In a E)) as [N|N];
   [apply In_NoDup_Permutation in N; desf;
    revert H0; rewrite !(Permutation_rsum (Permutation_map _ N)); ins|];
  rewrite map_upd_irr, ?upds, ?res_plusA; ins.
  by rewrite res_plusAC, H0, res_plusC.
Qed. 

Hint Resolve compat_mon_upd.

Lemma rsum_app l l' :
  rsum (l ++ l') = rsum l +!+ rsum l'.
Proof.
  induction l; simpl rsum; rewrite ?res_plus_emp_l, ?res_plusA; congruence.
Qed.

Lemma valid_node_ext_helper :
  forall acts N amap rf L a 
    (V : valid_node acts N amap rf L a) b,
    is_rmw (amap b) = false ->
    valid_node acts (b :: N) amap rf L a.
Proof.
  red; unfold NW; intros.
  red in V; desf.
  exists r, I; repeat (split; try done); ins; desf; eauto.
Qed.

Lemma valid_node_ext_helper2 :
  forall acts N amap rf L a 
    (V : valid_node acts N amap rf L a) b,
    rf b <> Some a ->
    valid_node acts (b :: N) amap rf L a.
Proof.
  red; unfold NW; intros.
  red in V; desf.
  exists r, I; repeat (split; try done); ins; desf; eauto.
Qed.

Lemma guar_escr :
  forall r_pre r alpha r_sb r_rf 
    (G : guar r_pre r alpha r_sb r_rf) rF sigma
    (E : res_escr (r_sb +!+ r_rf +!+ rF) sigma)
    (D : r +!+ rF <> Rundef),
  res_escr (r +!+ rF) sigma.
Proof.
  ins; apply res_escr_plusI; ins; rewrite <- ?res_plusA in *.
  destruct alpha; ins; desf; ins; desf; try rewrite res_plus_emp_r in *;
   rewrite <- ?res_plusA in *; repeat (apply res_escr_plusE in E; desf); eauto;
   rewrite ?res_escr_upd, ?res_escr_mupd in *; eauto; clear - E D AG;
  by left; red in AG; desf; rewrite <- (res_escr_upd r l (p_at tau S')), <- REQ; 
     apply res_escr_plusI; eauto; intro X; rewrite X in *; destruct r.
Qed.


Lemma vghost_escr :
  forall r r' I 
    (V : valid_ghost r r' I) sigma,
    res_escr r' sigma <-> res_escr r sigma \/ exists rr, In (sigma, rr) I.
Proof.
  ins; red in V; desf; ins.
Qed.

Lemma compat_upd1 :
  forall amap sb rf L
    (CC : compat amap sb rf L) te r
    (LT : res_lt r (L te)),
  compat amap sb rf (upd L te r).
Proof.
  red; ins.
  destruct (classic (In te E)) as [P|];
    [eapply In_NoDup_Permutation in P; desf;
     rewrite (Permutation_rsum (Permutation_map _ P)); ins |];
    rewrite !map_upd_irr; ins; rupd; eauto.
  intro; eapply (CC E); eauto.
  rewrite (Permutation_rsum (Permutation_map _ P)); ins. 
  by red in LT; desf; rewrite <- LT, res_plusA, res_plusAC, H, res_plusC.
Qed.

Lemma compat_upd2 :
  forall amap sb rf L
    (CC : compat amap sb rf L) te1 te2 
    (S : te_src te1 = te_src te2)
    (T : te_tgt te1 = None)
    (NEQ : te1 <> te2) r1
    (LT1 : res_lt r1 (L te1)) r2
    (LT2 : res_lt (r1 +!+ r2) (L te1 +!+ L te2)),
  compat amap sb rf (upd (upd L te1 r1) te2 r2).
Proof.
  red; ins.
  destruct (classic (In te2 E)) as [P|].
  2: by rewrite map_upd_irr; ins; eapply compat_upd1; eauto.
  eapply In_NoDup_Permutation in P; desf.
  red in LT2; desf.
  destruct (classic (In te1 l')) as [P'|].
    intro; eapply (CC E); ins; revert H.
    eapply In_NoDup_Permutation in P'; desf; 
      eauto using Permutation_nodup, Permutation_sym, nodup_consD.
    rewrite !(Permutation_rsum (Permutation_map _ P)); simpl; try done; rupd. 
    rewrite !(Permutation_rsum (Permutation_map _ P')); simpl; try done; rupd.
    rewrite !map_upd_irr; ins. 
    2: by unfold not in *; eauto using Permutation_in, Permutation_sym, in_cons.
    by rewrite res_plusAC in *; 
       rewrite <- res_plusA, <- LT2, res_plusA, res_plusAC, !res_plusA, H, res_plusC.
  intro; eapply (CC (te1 :: E)); ins; desf; desf; eauto.
    by econstructor; eauto; 
       intro; destruct H; eapply (Permutation_in _ P) in H1; ins; desf.
    by rewrite S in *; eauto using Permutation_in, Permutation_sym, in_eq.
  revert H0;
  rewrite !(Permutation_rsum (Permutation_map _ P)); simpl; try done; rupd. 
  rewrite !map_upd_irr, <- res_plusA, <- LT2; ins.
  by rewrite res_plusA, res_plusAC, res_plusA, H0, <- res_plusA, res_plusC.
Qed.

Lemma compat_upd2_bot :
  forall amap sb rf L
    (CC : compat amap sb rf L) te1 te2 
    (T1 : te_tgt te1 = None)
    (T2 : te_tgt te2 = None)
    (HB : happens_before amap sb rf (te_src te1) (te_src te2))
    (NEQ : te1 <> te2) r1
    (LT1 : res_lt r1 (L te1)) r2
    (LT2 : res_lt (r1 +!+ r2) (L te1 +!+ L te2)),
  compat amap sb rf (upd (upd L te1 r1) te2 r2).
Proof.
  red; ins.
  destruct (classic (In te2 E)) as [P|].
  2: by rewrite map_upd_irr; ins; eapply compat_upd1; eauto.
  eapply In_NoDup_Permutation in P; desf.
  red in LT2; desf.
  destruct (classic (In te1 l')) as [P'|].
    intro; eapply (CC E); ins; revert H.
    eapply In_NoDup_Permutation in P'; desf; 
      eauto using Permutation_nodup, Permutation_sym, nodup_consD.
    rewrite !(Permutation_rsum (Permutation_map _ P)); simpl; try done; rupd. 
    rewrite !(Permutation_rsum (Permutation_map _ P')); simpl; try done; rupd.
    rewrite !map_upd_irr; ins. 
    2: by unfold not in *; eauto using Permutation_in, Permutation_sym, in_cons.
    by rewrite res_plusAC in *; 
       rewrite <- res_plusA, <- LT2, res_plusA, res_plusAC, !res_plusA, H, res_plusC.
  intro; eapply (CC (te1 :: E)); ins; desf; desf; eauto.
    by econstructor; eauto; 
       intro; destruct H; eapply (Permutation_in _ P) in H1; ins; desf.
    eapply IND with (te:=te2) (te':=te');
    eauto using Permutation_in, Permutation_sym, in_eq, in_cons.
    by rewrite mhbE in *; desf; eauto using hb_trans.
  revert H0;
  rewrite !(Permutation_rsum (Permutation_map _ P)); simpl; try done; rupd. 
  rewrite !map_upd_irr, <- res_plusA, <- LT2; ins.
  by rewrite res_plusA, res_plusAC, res_plusA, H0, <- res_plusA, res_plusC.
Qed.

Lemma compat_upd2_acc :
  forall amap sb rf L te1 r1 r
    (CC : compat amap sb rf (upd L te1 (r1 +!+ r))) te2 
    (S : te_src te1 = te_src te2)
    (T1: te_tgt te1 = None)
    (T2: te_tgt te2 = None)
    (NEQ : te1 <> te2) r2
    (L2 : r2 = L te2 +!+ r),
  compat amap sb rf (upd (upd L te1 r1) te2 r2).
Proof.
  red; ins.
  destruct (classic (In te2 E)) as [P|].
  2: by rewrite map_upd_irr; ins; erewrite <- updss; 
        eapply compat_upd1; eauto; rupd; vauto.
  eapply In_NoDup_Permutation in P; desf.
  destruct (classic (In te1 l')) as [P'|].
    intro; eapply (CC E); ins; revert H.
    eapply In_NoDup_Permutation in P'; desf; 
      eauto using Permutation_nodup, Permutation_sym, nodup_consD.
    rewrite !(Permutation_rsum (Permutation_map _ P)); simpl; try done; rupd; try congruence.
    rewrite !(Permutation_rsum (Permutation_map _ P')); simpl; try done; rupd.
    rewrite !map_upd_irr; ins. 
    2: by unfold not in *; eauto using Permutation_in, Permutation_sym, in_cons.
    by rewrite res_plusA in *; rewrite (res_plusAC r). 
  intro; eapply (CC (te1 :: E)); ins; desf; desf; eauto.
    by econstructor; eauto; 
       intro; destruct H; eapply (Permutation_in _ P) in H1; ins; desf.
    by rewrite S in *; eauto using Permutation_in, Permutation_sym, in_eq.
  revert H0;
  rewrite !(Permutation_rsum (Permutation_map _ P)); simpl; try done; rupd; try congruence.
  rewrite !map_upd_irr, !res_plusA; ins.
  by rewrite (res_plusAC (L te2)), H0, res_plusC.
Qed.

Lemma compat_upd3 :
  forall amap sb rf L te3 rr
    (CC : compat amap sb rf (upd L te3 rr)) te1 te2 
    (S : te_src te1 = te_src te2)
    (T : te_tgt te1 = None)
    (TS : te_tgt te2 = Some (te_src te3))
    (T' : te_tgt te3 = None)
    (HB :  happens_before amap sb rf (te_src te1) (te_src te3))
    (NEQ : te1 <> te2) 
    (NEQ' : te1 <> te3) r1
    (LT1 : res_lt r1 (L te1)) r2 r3
    (LT2 : res_lt (r1 +!+ r2) (L te1 +!+ L te2))
    (LT3 : res_lt (r1 +!+ r3) (L te1 +!+ rr)),
  compat amap sb rf (upd (upd (upd L te1 r1) te2 r2) te3 r3).
Proof.
  red; ins.
  assert (NEQ'' : te2 <> te3) by congruence.
  destruct (classic (In te3 E)) as [P|].
  2: erewrite <- updss, map_upd_irr; ins. 
  2: by rewrite !(updC (l':=te3)); ins; eapply compat_upd2; eauto; rupd.
  assert (~ In te2 E) by (intro X; eapply (IND _ P _ X); eauto; vauto).
  rewrite updC, map_upd_irr; ins.
  clear te2 r2 LT2 H S TS NEQ NEQ''.  
  erewrite <- updss, (updC (l:=te1)); ins.
  eapply compat_upd2_bot; eauto; rupd.
Qed.

Lemma in_rf_upds acts L a b r :
  NoDup acts ->
  In a acts ->
  in_rf acts L b = res_emp ->
  in_rf acts (upd L (Trf a b) r) b = r.
Proof.
  unfold in_rf; intros ND H.
  eapply In_NoDup_Permutation in H; desf.
  rewrite !(Permutation_rsum (Permutation_map _ (Permutation_map _ H))); simpl; rupd.
  rewrite map_upd_irr; try (by rewrite in_map_iff; intro; desf).
  by intro X; apply res_plus_eq_empD, proj2 in X; rewrite X, res_plus_emp_r.
Qed.


Lemma no_rmw_between :
  forall b acts amap sb rf mo
    (CONS : Consistent (b :: acts) amap sb rf mo) l v
    (H : amap b = Astore ATat l v) a
    (MO : immediate mo a b) c
    (IN: In c acts)
    (RF: rf c = Some a)
    (RMW: is_rmw (amap c)),
    False.
Proof.
  ins.
  red in CONS; desf.
  exploit Crmw; eauto.
  unfold immediate in *; ins; desf.
  specialize (Crf c); rewrite RF in *; desf.
  red in Cmo; desf.
  destruct (eqP b c); desf; [by rewrite H in *|].
  specialize (Cmo _ _ MO); desf; rewrite H in *; ins; desf.
  assert (l1 = l0) by (destruct (amap a); ins; desf); subst.
  specialize (Cmo1 l0 b); ins; rewrite H in *; destruct (Cmo1 eq_refl c); eauto.
  by destruct (amap c); ins; desf.
Qed.

Lemma helper_plus_insb_rfS_def :
  forall a b acts amap sb rf mo L
    (CONS: Consistent (b::acts) amap sb rf mo)
    (ND: NoDup (b :: acts))
    (PC : prefix_closed sb rf acts)
    (CC: compat amap sb rf L)
    (IN: In a acts) 
    (NIN: ~ In b acts)
    (LR : label_restr (b::acts) amap sb rf L),
  in_sb (b :: acts) L b +!+ L (TrfS a) <> Rundef.
Proof.
  ins; assert (NEQ: a <> b) by congruence.
  replace (in_sb (b :: acts) L b) with 
    (rsum (map L (map (Tsb^~ b) (filter (mydecf (sb^~ b)) acts)))). 
  unfold in_sb; rewrite res_plusC.
  intro; eapply (CC (TrfS a :: map (Tsb^~ b) (filter (mydecf (sb^~ b)) acts))).
      constructor; [by rewrite in_map_iff; intro; desf|].
      by eapply nodup_map; try eapply nodup_filter; ins; try solve [by inv ND|congruence]. 
    by ins; desf; ins; try rewrite in_map_iff, mhbE in *; desf; ins; desf; 
       try rewrite filter_In in *; desf; eauto using prefix_closed_hb.
    by ins; rewrite in_map_iff; intro; desf.
  unfold in_sb; ins.  
  rewrite (proj1 LR), res_plus_emp_l, rsum_map_map_filter; auto.
    by ins; apply LR; ins.
  by intro; red in CONS; desc; edestruct ACYC; vauto.
Qed. 

Lemma helper_plus_inc_rfS_def :
  forall a b acts amap sb rf mo L
    (CONS: Consistent (b::acts) amap sb rf mo)
    (ND: NoDup (b :: acts))
    (PC : prefix_closed sb rf acts)
    (CC: compat amap sb rf L)
    (IN: In a acts) 
    (NIN: ~ In b acts)
    (LR : label_restr (b::acts) amap sb rf L),
  incoming (b :: acts) L b +!+ L (TrfS a) <> Rundef.
Proof.
  ins; assert (NEQ: a <> b) by congruence.
  intro; eapply (CC (TrfS a :: map (Tsb^~ b) acts ++ map (Trf^~ b) acts ++ map (Tesc^~ b) acts)).
      constructor; [by rewrite ?in_app_iff, ?in_map_iff; intro; desf|].
      inv ND; rewrite !nodup_app; intuition; try eapply nodup_map; eauto; ins; try congruence;
      unfold disjoint; ins; rewrite ?in_app_iff, ?in_map_iff in *; desf.
    by ins; desf; ins; rewrite ?in_app_iff, ?in_map_iff, mhbE in *; desf; ins; desf; 
       try rewrite filter_In in *; desf; eauto using prefix_closed_hb.
  ins; rewrite !map_app, !rsum_app.
  rewrite res_plusC, <- H; unfold incoming, in_sb, in_rf, in_esc; ins.
  red in CONS; desc.
  rewrite (proj1 LR), (proj1 (proj2 LR)), (proj1 (proj2 (proj2 LR))), !res_plus_emp_l; ins;
  intro; edestruct ACYC; vauto; eauto using hb_in_ct.
Qed. 

Lemma helper_escr_ffun L I r b pi rem acts :
  pi = L (TcondS b) +!+ rem ->
  pi <> Rundef ->
  res_lt (rsum (map snd (filter (escr_ffun L acts) I))) r ->
  res_lt (rsum (map snd (filter (escr_ffun (upd L (TcondS b) pi) acts) I))) r.
Proof.
  intros; eapply res_lt_trans; eauto; clear r H1; subst.
  induction I; ins; eauto using res_lt_refl.
  unfold escr_ffun at 1 4; desf; ins; desf; eauto using res_lt_frame_l, res_lt_plus_l.
  destruct n; repeat eexists; eauto; ins.
  eapply (e1 c); unfold upd; desf; eauto using RP_ext.
Qed. 

Lemma vghost_ext :
  forall r r' I r_pre l
    (VG: valid_ghost r r' I)
    (EXT: prot_trans_ok (res_get r l) (res_get r_pre l)),
    prot_trans_ok (res_get r' l) (res_get r_pre l).
Proof.
  unfold valid_ghost; ins; desf.
Qed.

Lemma vghost_env_move :
  forall r r' I r'' l v
    (VG: valid_ghost r r' I)
    (EXT: env_move r l v r'')
    (DEF: res_plus r' r'' <> Rundef),
    env_move r' l v r''.
Proof.
  unfold valid_ghost, env_move; ins; desf; repeat eexists; try edone.
Qed.

Lemma safe_mon n e QQ r m : safe n e QQ r -> m <= n -> safe m e QQ r.
Proof.
  induction[m e QQ r] n; ins; desf; destruct m; ins; unfold NW; repeat split; ins; desf;
    try (exfalso; omega); apply le_S_n in H0.
    exploit FORK; eauto; intro; desf; eauto 10. 
  exploit STEP; eauto; intro; desf. 
  eexists,;eauto; ins; exploit x1; eauto; desf; intro; desf; eauto.
Qed.


Ltac discharge_eq_emp_tac :=
 repeat first [ apply rsum_map_map_eq_empI; intros ? _; ins; desf; ins
              | apply res_plus_eq_empI ]; ins; desf; desf; 
  try match goal with H: ~ (_ \/ _) |- _ => solve [destruct H; vauto] end.


Lemma replace_out_sb :
  forall a b c L r1 r2 acts (ND: NoDup acts) (IN : In b acts)
    (EMP: L (Tsb a b) +!+ L (TsbS a) = r1 +!+ r2),
  out_sb acts (upd (upd L (Tsb a b) r1) (TsbS a) r2) c = out_sb acts L c.
Proof.
  ins; unfold out_sb.
  destruct (classic (a = c)); desf; rupd; try congruence.
  2: by rewrite !map_upd_irr; ins; try by rewrite in_map_iff; intro; desf.
  eapply In_NoDup_Permutation in IN; eauto; desf.
  rewrite !(Permutation_rsum (Permutation_map _ (Permutation_map _ IN))); ins; rupd.
  rewrite !map_upd_irr; ins; try by rewrite in_map_iff; intro; desf.
  by rewrite <- !res_plusA; f_equal; rewrite res_plusC, <- EMP, res_plusC.
Qed.  

Lemma rsum_eq_empI l :
  (forall x, In x l -> x = res_emp) -> rsum l = res_emp.
Proof.
  induction l; ins; rewrite (H a), res_plus_emp_l; eauto.
Qed.  

Lemma Consistent_sw_in :
  forall acts amap sb rf mo (C : Consistent acts amap sb rf mo) a b,
  synchronizes_with amap rf a b ->
  In a acts /\ In b acts.
Proof.
  by destruct 2 as [_[??]]; split; apply (proj1 (proj1 C)); intro X; rewrite X in *.
Qed. 

Lemma Consistent_hb_in :
  forall acts amap sb rf mo (C : Consistent acts amap sb rf mo) a b,
  happens_before amap sb rf a b ->
  In a acts /\ In b acts.
Proof.
  split; induction H; eauto; unfold imm_happens_before in *; desf;
  try eapply (proj2 (proj1 C)) in H;
  try eapply (Consistent_sw_in C) in H; desf.
Qed. 

Lemma vghost_get_at :
  forall r1 r r' I 
    (VG : valid_ghost r1 (r +!+ r') I)
    (RS : res_lt (res_strip r1) r) l tau S
    (G1 : res_get r1 l = p_at tau S),
    res_get r l = p_at tau S.
Proof.
  ins; red in VG; desf; clear I VG1; red in RS; desf; ins; desf; ins; desf.
  inv_lplus l; rewrite G1 in *; desf; ins; desf.
  rewrite <- H0 in *; f_equal.
  ins; desf.
  apply slist_eq2; ins.
  by split; [clear Heq4|clear Heq5]; unfold mk_slist in *; desf; ins;
     rewrite in_undup_iff, In_sort, in_app_iff; desf; eauto.
Qed.


Lemma vghost_get_back_at :
  forall r1 r r' I 
    (VG : valid_ghost r1 (r +!+ r') I)
    (RS : res_lt (res_strip r1) r) l tau S
    (G1 : res_get r l = p_at tau S),
    res_get r1 l = p_at tau S.
Proof.
  ins; red in VG; desf; clear I VG1; red in RS; desf; ins; desf; ins; desf.
  inv_lplus l; rewrite G1 in *; desf; ins; desf.
  rewrite <- H0 in *; f_equal.
  ins; desf.
  apply slist_eq2; ins.
  by split; [clear Heq4|clear Heq5]; unfold mk_slist in *; desf; ins;
     rewrite in_undup_iff, In_sort, in_app_iff; desf; eauto.
Qed.

Lemma vghost_get_back :
  forall r1 r2 r r' I 
    (VG : valid_ghost (r1 +!+ r2) (r +!+ r') I) l tau S
    (G  : res_get r l = p_at tau S)
    (NBOT : res_get r1 l <> p_bot),
    exists S_pre, 
      res_get r1 l = p_at tau S_pre.
Proof.
  ins; red in VG; desf; clear I VG1. 
  destruct r1, r; ins; desf; ins; desf.
  inv_lplus l; rewrite G in *; destruct (pmap l); ins; desf; 
  rewrite <- H0 in *; desf; eauto.
Qed.

Lemma conform_upd_irr :
  forall acts amap mo L acts', 
    conform acts amap mo L acts' ->
    forall b r,
      conform acts amap mo (upd L (TsbS b) r) acts'.
Proof.
  unfold conform, out_rf; intros; rewrite !map_upd_irr; rupd; auto;
  rewrite in_map_iff; intro; desf.
Qed.

Lemma conform_ext_irr :
  forall acts amap sb rf mo L acts', 
    Consistent acts amap sb rf mo ->
    conform acts amap mo L acts' ->
    forall b,
      (forall l, ~ is_writeL (amap b) l) ->
      conform acts amap mo L (b::acts').
Proof.
  unfold conform; intros.
  red in H; desf. 
  assert (M := proj1 Cmo _ _ MO); desf.
  ins; desf; eauto; try by edestruct H1; eauto.
Qed.

Lemma valid_node_upd_irr acts amap rf L te a r : 
  a <> te_src te ->
  te_tgt te = None ->
  (forall a : actid, te <> TcondS a) ->
  valid_node acts acts amap rf L a ->
  valid_node acts acts amap rf (upd L te r) a.
Proof.
  intros; red in H2; desf; red; unfold NW.
  eexists r0, I.
  unfold incoming, in_sb, in_rf, in_esc, out_esc, out_rf, out_sb in *.
  rewrite !map_upd_irr, escr_ffun_upd_irr; rupd; try (by ins; intro; desf);
    try (by rewrite in_map_iff; intro; desf).
  repeat (split; ins; rupd); eauto; try (by ins; intro; desf).
Qed.

Lemma compat_guar1 amap sb rf L b r :
  compat amap sb rf (upd L (TescS b) (L (TescS b) +!+ r)) ->
  compat amap sb rf (upd L (TsbS b) r).
Proof. 
  intros.
  erewrite <- (updI L (TescS b)), <- (updss L).
  eapply compat_upd2; eauto; ins; rupd; vauto.
Qed.

Lemma compat_guar2 amap sb rf L te r l pi :
  (res_get r l = p_uninit \/ exists v, res_get r l = p_na v) ->
  compat amap sb rf (upd L te r) ->
  compat amap sb rf (upd L te (res_upd r l pi)).
Proof. 
  repeat red; intros.
  exploit H0; eauto.
  revert H1.
  destruct (classic (In te E)) as [P|];
    [eapply In_NoDup_Permutation in P; desc; clarify;
     rewrite !(Permutation_rsum (Permutation_map _ P)); try done |]; simpl;
    rewrite !map_upd_irr; try done; rupd; eauto.
  destruct r; ins; desf;
  eapply lift_plus_none_inv in Heq2; desc; inv_lplus l0; desf; desf;
  rewrite H in *; destruct pi; ins; desf.
Qed.

Lemma vghost_get_back_uninit :
  forall r1 r r' I 
    (VG : valid_ghost r1 (r +!+ r') I) l
    (G  : res_get r l = p_uninit),
  res_get r1 l = p_uninit. 
Proof.
  ins; red in VG; desf; clear I VG1. 
  destruct r; ins; desf; ins; desf.
  inv_lplus l; rewrite G in *; ins; desf.
Qed.

Lemma helper_plus_inc_escS_def :
  forall a b acts amap sb rf mo L
    (CONS: Consistent (b::acts) amap sb rf mo)
    (ND: NoDup (b :: acts))
    (PC : prefix_closed sb rf acts)
    (CC: compat amap sb rf L)
    (IN: In a acts) 
    (NIN: ~ In b acts)
    (LR : label_restr (b::acts) amap sb rf L),
  incoming (b :: acts) L b +!+ L (TescS a) <> Rundef.
Proof.
  ins; assert (NEQ: a <> b) by congruence.
  intro; eapply (CC (TescS a :: map (Tsb^~ b) acts ++ map (Trf^~ b) acts ++ map (Tesc^~ b) acts)).
      constructor; [by rewrite ?in_app_iff, ?in_map_iff; intro; desf|].
      inv ND; rewrite !nodup_app; intuition; try eapply nodup_map; eauto; ins; try congruence;
      unfold disjoint; ins; rewrite ?in_app_iff, ?in_map_iff in *; desf.
    by ins; desf; ins; rewrite ?in_app_iff, ?in_map_iff, mhbE in *; desf; ins; desf; 
       try rewrite filter_In in *; desf; eauto using prefix_closed_hb.
  ins; rewrite !map_app, !rsum_app.
  rewrite res_plusC, <- H; unfold incoming, in_sb, in_rf, in_esc; ins.
  red in CONS; desc.
  rewrite (proj1 LR), (proj1 (proj2 LR)), (proj1 (proj2 (proj2 LR))), !res_plus_emp_l; ins;
  intro; edestruct ACYC; vauto; eauto using hb_in_ct.
Qed. 

Lemma incoming_upd_irr : 
  forall acts L te r a,
    te_tgt te = None ->
    incoming acts (upd L te r) a = 
    incoming acts L a. 
Proof.
  unfold incoming, in_sb, in_rf, in_esc; ins.
  rewrite !map_upd_irr; ins; rewrite in_map_iff; intro; desf.
Qed.

Lemma is_atomic_write_weaken typ l : is_atomic_writeL typ l -> is_writeL typ l.
Proof. destruct typ; ins; desf. Qed.

Hint Immediate is_atomic_write_weaken.

Lemma prot_trans_ok_trans x y z :
  prot_trans_ok x y ->
  prot_trans_ok y z ->
  prot_trans_ok x z.
Proof.
  destruct x; ins; desf; desf; ins; desf; eauto using slist_lt_trans. 
Qed.

Lemma helper_plus_inc_escS_back_def :
  forall a b acts amap sb rf mo L
    (CONS: Consistent acts amap sb rf mo)
    (ND: NoDup acts)
    (CC: compat amap sb rf L)
    (MO: mo b a)
    (LR : label_restr acts amap sb rf L),
  incoming acts L a +!+ L (TescS b) <> Rundef.
Proof. 
  ins (*; assert (NEQ: a <> b) by congruence*).
  intro; eapply (CC (TescS b :: 
      map (Tsb^~ a) (filter (mydecf (fun c => happens_before amap sb rf c a)) (acts)) ++ 
      map (Trf^~ a) (filter (mydecf (fun c => happens_before amap sb rf c a)) (acts)) ++ 
      map (Tesc^~ a) (filter (mydecf (fun c => happens_before amap sb rf c a)) (acts)))).
      constructor; [by rewrite ?in_app_iff, ?in_map_iff; intro; desf|].
      rewrite !nodup_app; intuition; try eapply nodup_map; eauto; ins; try congruence;
      unfold disjoint; ins; rewrite ?in_app_iff, ?in_map_iff in *; desf; eauto using nodup_filter.
    red in CONS; desf.
    ins.
    assert (b0 = a).
      by clear IN; ins; desf; ins; rewrite ?in_app_iff, ?in_map_iff in *; desf; ins; desf.
    desf; ins.
      red in Cmo; desc; rewrite mhbE in *; desf; destruct (Cmo _ _ MO) as (?&?&?).
        eby eapply Cmo2.
      by eapply Cmo2, Cmo0, Cmo3; try eapply HB; eauto.
    assert (HB': happens_before amap sb rf (te_src te) a).
      by clear IN'; ins; desf; ins; rewrite ?in_app_iff, ?in_map_iff in *; desf; ins; desf;
         rewrite ?filter_In in *; desf; unfold mydecf in *; desf.
    by clear IN; desf; rewrite ?in_app_iff, ?in_map_iff, mhbE in *; desf; ins; desf;
       try rewrite filter_In in *; desf; unfold mydecf in *; desf;
       edestruct ACYC; eauto using hb_trans, hb_in_ct.
  red in CONS; desc; ins.
  rewrite res_plusC, !map_app, !rsum_app, <- H; unfold incoming, in_sb, in_rf, in_esc; ins.
  by rewrite !rsum_map_map_filter; ins; apply LR; intro; desf; eauto. 
Qed.

Lemma plus_upd_def :
  forall r l tau S
    (G : res_get r l = p_at tau S) r' S'
    (P : res_upd r l (p_at tau S') +!+ r' = Rundef)
    (IMP: forall S'' (G' : res_get r' l = p_at tau S'')
           s' (IN': In s' (SL_list S')) (NIN': ~ In s' (SL_list S)) 
           s'' (IN'': In s'' (SL_list S'')), prot_trans tau s' s'' \/ prot_trans tau s'' s'),
  r +!+ r' = Rundef.
Proof.
  ins; destruct r; ins; desf.
  eapply lift_plus_none_inv in Heq1; desc; inv_lplus l0. 
  desf; desf; rewrite G in *; ins; desf.
  unfold mk_slist in *; desf; destruct n; clear H0; desf.
  split; ins.
   by destruct S' as [[]].
  specialize (IMP _ eq_refl); clear Heq0 Heq2 G.
  destruct s, S', S; ins.

  rewrite in_app_iff in *; desf; eauto using in_or_app;
    try solve [eapply Sorted_total; try exact H; unfold reflexive; eauto using prot_trans_refl]. 
    destruct (classic (In y SL_list1)); eauto 8 using in_or_app.
    exploit IMP; eauto; tauto.
  destruct (classic (In x SL_list1)); eauto 8 using in_or_app.
Qed. 

Lemma get_plus_at_oneD :
  forall r r' l tau S s, 
    res_get (res_plus r r') l = p_at tau S ->
    In s (SL_list S) ->
    exists S', (res_get r l = p_at tau S' \/ res_get r' l = p_at tau S') /\ In s (SL_list S').
Proof.
  destruct r, r'; ins; desf; ins.
  inv_lplus l; unfold prot_plus in *; rewrite H in *; desf; eauto.
  unfold mk_slist in *; desf; ins.
  rewrite in_undup_iff, In_sort, in_app_iff in *; desf; eauto.
Qed.

Lemma get_rsum_at_oneD :
  forall rl l tau S s, 
    res_get (rsum rl) l = p_at tau S ->
    In s (SL_list S) ->
    exists r S', In r rl /\ res_get r l = p_at tau S' /\ In s (SL_list S').
Proof.
  induction rl; ins.
  eapply get_plus_at_oneD in H; eauto; desf; eauto 8.
  exploit IHrl; eauto; intro; desf; eauto 8.
Qed.

Lemma valid_ghost_refl r : r <> Rundef -> valid_ghost r r nil.
Proof.
  red; destruct r; ins; repeat eexists; ins; desf; eauto.
Qed.

Lemma rely_def r act r' : rely r act r' -> r' <> Rundef.
Proof.
  destruct act; ins; desf; desf.
Qed.

Lemma filter_eq_nil A (f : A -> bool) l :
  (forall x, In x l -> f x = false) -> filter f l = nil.
Proof.
  induction l; ins; rewrite H; eauto.
Qed.

Lemma hb_in_acts2 acts amap sb rf mo a b :
  Consistent acts amap sb rf mo ->
  happens_before amap sb rf a b ->
  In b acts.
Proof.
  induction 2 as [? ? [|[? ?]]|]; eauto; red in H; desc. 
    by apply (proj2 FIN), proj2 in H0.
  by apply (proj1 FIN); intro X; rewrite X in *.
Qed.


