(******************************************************************************)
(*   #####   #####    ####      Formalization of the GPS program logic        *)
(*  ##       ##  ##  ##         by Turon, Vafeiadis and Dreyer                *)
(*  ##  ###  #####    ####                                                    *)
(*  ##   ##  ##          ##     Copyright (c) Viktor Vafeiadis, 2013          *)
(*   #####   ##       ####      See README.txt for license                    *)
(******************************************************************************)

Require Import Vbase Relations ClassicalDescription ClassicalChoice.
Require Import List Permutation Sorting extralib. 
Require Import c11 exps.
Require Import GpsModel GpsModelLemmas.

Set Implicit Arguments.
Local Open Scope string_scope.

(******************************************************************************)
(******************************************************************************)
(** * 2.4. Syntax and semantics *)
(******************************************************************************)
(******************************************************************************)


(******************************************************************************)
(** ** 2.4.1. Parameters *)
(******************************************************************************)


(******************************************************************************)
(** ** 2.4.2. Syntax *)
(******************************************************************************)

Inductive assn :=
  | Pure (P : Prop) 
  | Base (P : RProp)
  | Conj (P P' : assn) 
  | Disj (P P' : assn)
  | Impl (P P' : assn)
  | Forall A (PP: A -> assn)
  | Exists A (PP: A -> assn)
  | Star (P P' : assn)
  | Box  (P : assn)
  | Uninit (l : nat) 
  | Pt   (l v : nat)
  | At   (l : nat) (s: state) (tau : protTy)
  | Gres (gres : ghost_resource)
  | Escr (escr : escrowTy).

 
(******************************************************************************)
(** ** 2.4.3. Proposition semantics *)
(******************************************************************************)

Program Definition Rat l s tau :=
  {| RP_car := fun r => exists S, res_get r l = p_at tau S /\ In s (SL_list S) |}.
Next Obligation. red; ins; desf. Qed.
Next Obligation. exploit res_get_plus; eauto; ins; desf; eauto. Qed.

Fixpoint assn_sem (P : assn) { struct P } :=
  match P with 
    | Pure P => Rpure P
    | Base P => P
    | Conj P P' => Rconj (assn_sem P) (assn_sem P')
    | Disj P P' => Rdisj (assn_sem P) (assn_sem P')
    | Impl P P' => Rimpl (assn_sem P) (assn_sem P')
    | Forall A PP => Rforall (fun x => assn_sem (PP x)) 
    | Exists A PP => Rexists (fun x => assn_sem (PP x))
    | Star P P' => Rstar (assn_sem P) (assn_sem P')
    | Box  P   => Rbox (assn_sem P)
    | Uninit l => Runinit l 
    | Pt l v => Rpt l v 
    | At l s tau => Rat l s tau
    | Gres g => Rgres g
    | Escr sigma => Rescr sigma
  end.

(******************************************************************************)
(** ** 2.4.4. Ghost move semantics *)
(******************************************************************************)

Definition gimp (P Q : RProp) := forall r (SAT: P r), ghost_move r Q.

(******************************************************************************)
(** ** 2.4.5. Hoare triple semantics *)
(******************************************************************************)

Definition triple P (e : exp) QQ :=
  exists P', 
    gimp P P' /\
    forall r' (SAT': P' r') n,
      safe n e (fun n => QQ n) r'.


(******************************************************************************)
(******************************************************************************)
(** * 2.5. Proof theory *)
(******************************************************************************)
(******************************************************************************)

Lemma RProp_eq (P Q : RProp) :
  (forall r, P r <-> Q r) -> P = Q.
Proof.
  ins; cut (RP_car P = RP_car Q).
    by destruct P, Q; ins; subst; f_equal; apply proof_irrelevance.  
  by extensionality r; apply prop_ext. 
Qed.

Lemma Rstar_true P : Rstar P Rtrue = P.
Proof.
  apply RProp_eq; intro r; ins; destruct P; split; ins; desf; eauto.
  split; try red; ins; desf.
  by eexists r, res_emp; rewrite res_plus_emp_r.
Qed.

Lemma RstarA P Q R : Rstar (Rstar P Q) R = Rstar P (Rstar Q R).
Proof.
  ins; apply RProp_eq; split; ins; desf; split; eauto.
    rewrite res_plusA in *; repeat eexists; eauto.
    by intro X; rewrite X in *; destruct r0; ins.
  rewrite <- res_plusA in *; repeat eexists; eauto.
  by intro X; rewrite X in *.
Qed.

Lemma RstarC P Q : Rstar P Q = Rstar Q P.
Proof.
  ins; apply RProp_eq; split; ins; desf; split; rewrite res_plusC in *;
  repeat eexists; eauto.
Qed.

Lemma RstarAC P Q R : Rstar P (Rstar Q R) = Rstar Q (Rstar P R).
Proof.
  rewrite <- !RstarA; f_equal; apply RstarC.
Qed.

Lemma Rstar_singl a b :
  res_plus a b <> Rundef ->
  Rstar (Rsingl a) (Rsingl b) = Rsingl (res_plus a b).
Proof.
  ins; apply RProp_eq; intro r; ins; split; ins; desf; split; ins.
  exists (res_plus r'0 r').
  rewrite !res_plusA; f_equal; rewrite <- !res_plusA; f_equal; apply res_plusC.
  rewrite res_plusA in *.
  eexists a, (res_plus b r'); repeat split; try red; ins; desf; eauto using res_plus_emp_r.
  rewrite H1 in *; destruct a; ins.
Qed.

Lemma guar_def r_pre r' alpha r rx :
  guar r_pre r' alpha r rx ->
  r <> Rundef /\ rx <> Rundef.
Proof.
  destruct alpha; ins; desf; desf; destruct r'; ins;
  unfold atomic_guar in *; desf; destruct r, rx; ins.
Qed.

Lemma res_get_plus_na r r' (D: res_plus r r' <> Rundef) l v :
  res_get (res_plus r r') l = p_na v <->
  res_get r l = p_na v \/ res_get r' l = p_na v. 
Proof.
  destruct r; ins; desf; ins; inv_lplus l; unfold prot_plus in *; desf; vauto;
  split; ins; desf;
    try solve [left; congruence | right; congruence | congruence].
Qed. 

Lemma Rstar_at l s tau P r :
  Rstar (Rat l s tau) P r <-> 
  P r /\ exists S, res_get r l = p_at tau S /\ In s (SL_list S).
Proof.
  split; ins; desf; split; ins.
    by rewrite res_plusC in *; eauto using RP_ext. 
    by exploit res_get_plus; eauto; ins; desf; eauto.
    by intro; desf.
  eexists (res_strip r),r; rewrite res_plus_strip_l; repeat eexists; eauto.
  by destruct r; ins; rewrite H0.
Qed.

Lemma RP_strip_ext (P : RProp) r r' : 
  P (res_strip r) ->
  res_plus r r' <> Rundef ->
  P (res_strip (res_plus r r')).
Proof.
  ins; rewrite res_strip_plus; ins.
  eapply RP_ext; eauto using res_plus_strip2_def.
Qed.

Lemma RP_ext_upd (P : RProp) r l pi : 
  P r -> 
  prot_plus (res_get r l) pi = Some pi ->
  P (res_upd r l pi).  
Proof.
  ins; cut (res_upd r l pi = res_plus r (res_upd res_emp l pi)).
    by intro X; rewrite X; eapply RP_ext; eauto; rewrite <- X; destruct r, P.
  rewrite prot_plusC, res_plusC in *; destruct r; ins; desf;
  rewrite ?gres_plusI in *; desf; f_equal;
    try (by extensionality sigma; apply prop_ext; tauto).

  extensionality l'; eapply lift_plus_some_inv with (l:=l') in Heq.
  destruct (eqnP l' l); ins; desf; ins; desf; destruct pi; ins; desf. 

  exfalso; eapply lift_plus_none_inv in Heq; desc.
  destruct (eqnP l0 l); ins; desf; ins; desf; destruct pi; ins; desf. 
Qed.

(******************************************************************************)
(** ** 2.5.1. Necessitation *)
(******************************************************************************)

Definition implies (P Q : RProp) := forall r (SAT: P r), Q r.

Lemma Rbox1 P : implies (Rbox P) P. 
Proof.
  red; ins; rewrite <- res_plus_strip_idemp, res_plusC; apply RP_ext; ins.
  rewrite res_plusC, res_plus_strip_idemp; intro; destruct P; ins; desf.
Qed.

Lemma Rbox2 P : implies (Rbox P) (Rbox (Rbox P)). 
Proof.
  by red; ins; rewrite res_strip_idemp.
Qed.

Lemma Rstar_box P Q r :
  Rstar P (Rbox Q) r <-> P r /\ Q (res_strip r). 
Proof.
  split; ins; desf; split; ins; eauto using RP_ext.
  rewrite res_plusC in *.
  rewrite res_strip_plus; ins; apply RP_ext; eauto. 
  rewrite <- res_strip_plus; ins. 
   destruct (res_plus r2 r1); ins.
  destruct Q; intro; desf.
  eexists r, (res_strip r); repeat split; eauto using eq_sym, res_plus_strip_idemp.
  by rewrite res_strip_idemp. 
Qed.

Lemma Rbox_pure P :
  implies (Rpure P) (Rbox (Rpure P)).
Proof.
  red; ins; desf; destruct r; ins.
Qed.

Lemma Rbox_at l s tau : 
  implies (Rat l s tau) (Rbox (Rat l s tau)).
Proof.
  red; ins; desf; repeat eexists; eauto.
  by destruct r; ins; rewrite SAT.
Qed.

Lemma Rbox_escr sigma : 
  implies (assn_sem (Escr sigma)) 
          (Rbox (assn_sem (Escr sigma))).
Proof.
  red; ins; desf; destruct r; ins.
Qed.

(******************************************************************************)
(** ** 2.5.2. Separation *)
(******************************************************************************)

Lemma Rstar_gres1 g g' r :
  Rstar (Rgres g) (Rgres g') r <->
  exists g'', gres_plus g g' = Some g'' /\
              (assn_sem (Gres g'')) r.
Proof.
  destruct (gres_plus g g') as [g''|] eqn: X.
  unfold Rgres; rewrite Rstar_singl; ins;
  rewrite lift_plus_emp_l in *; desf; repeat split; ins; desf; try done;
  rewrite lift_plus_emp_l in *; desf; [eexists,,; try done|]; 
  eexists (Rdef p gres escr); simpl;
  rewrite lift_plus_emp_l in *; desf; f_equal;
  extensionality l; apply prop_ext; tauto.

  split; intros; desf.
  exfalso; hnf in H; desf.
  hnf in H0, H1; desf; clear H0 H1.
  rewrite !res_plusA, res_plusAC, res_plusC, <- res_plusA in H.
  ins; desf.
Qed.

Lemma Rstar_at1 l s tau s' tau' : 
  implies (Rstar (Rat l s tau) (Rat l s' tau'))
          (Rpure (tau = tau' /\ (prot_trans tau s s' \/ prot_trans tau s' s))).
Proof.
  red; ins; desf; split; try done.
  destruct r1, r2; ins; desf.
  inv_lplus l; unfold prot_plus, mk_slist in *; desf; clear H0; desf.
  split; ins; eapply a0; eauto using in_or_app.
Qed.


(******************************************************************************)
(** ** 2.5.3. Ghost moves *)
(******************************************************************************)

Lemma GM_singl r : r <> Rundef -> ghost_move r (Rsingl r). 
Proof. ins; apply GM_base; split; eauto using res_plus_emp_r. Qed.

Lemma gimp_refl P : gimp P P.
Proof. red; eauto using ghost_move. Qed.

Lemma gimp_implies P Q : implies P Q -> gimp P Q.
Proof. unfold implies, gimp; eauto using ghost_move. Qed.

Lemma gimp_trans : transitive _ gimp.
Proof. repeat red; ins; eapply GM_trans; eauto. Qed.

Lemma GM_frame r1 r2 (Q F : RProp) :
  ghost_move r1 Q ->
  F r2 ->
  res_plus r1 r2 <> Rundef ->
  ghost_move (res_plus r1 r2) (Rstar Q F).
Proof.
  intros until 1; revert r2; induction H; ins; desf.
  econstructor 1; ins; desf; eauto 7. 

  econstructor 2; eauto; ins; desf; eauto. 

  econstructor 3; eauto; unfold NW; ins; desf.
  rewrite res_plusA, (res_plusC r2), <- res_plusA in *; desf.
  repeat eexists; eauto; eapply PEQ; eauto.
  by intro X; rewrite X in *.

  ins; desf.
  destruct (gres_plus g' gres) as [g1|] eqn: M.
    2: by rewrite <- (GDEF gres) in *; desf. 
  econstructor 4 with (g':=g1); repeat eexists; ins; eauto; ins; desf. 
    case_eq (gres_plus g1 gF) as Y; ins. 
    destruct (gres_plusA M Y) as (gg & ? & ?).
    destruct (gres_plusA2 Heq0 H1); desf; rewrite GDEF in *; congruence.
  case_eq (gres_plus g0 gF) as Y; ins. 
  destruct (gres_plusA Heq0 Y) as (gg & ? & ?).
  destruct (gres_plusA2 M H1); desf; rewrite <- GDEF in *; congruence.

  rewrite res_plusA, (res_plusC r2), <- res_plusA in *.
  econstructor 5; try edone; repeat (eexists; try edone); eauto. unfold NW; ins; desf.
    destruct (res_plus r1 r0); ins; clear.
    by rewrite lift_plusC, lift_plus_emp_l, gres_plusC, gres_plusI;
       ins; auto using prot_plusC.
  by rewrite res_plusC, res_plusAC, res_plusA.

  rewrite res_plusA in *.
  econstructor 6; eauto; ins; desf. 
    by rewrite <- res_plusA in *; apply res_escr_plus; eauto.
  rewrite !res_plusA, res_plusAC, res_plusC in H1 |- *.
  exploit (IN' (res_plus r0 (res_plus r' r3))); eauto;
  split; repeat eexists; eauto using res_plusA.
    by intro X; rewrite X in *.
  by rewrite <- !res_plusA in *; intro X; rewrite X in *. 
Qed.

Lemma gimp_frame P Q F :
  gimp P Q ->
  gimp (Rstar P F) (Rstar Q F).
Proof.
  unfold gimp; ins; desf; eapply GM_frame; eauto. 
Qed.

Lemma gimp_star P Q P' Q' :
  gimp P Q ->
  gimp P' Q' ->
  gimp (Rstar P P') (Rstar Q Q').
Proof.
  ins; eapply gimp_trans; eauto using gimp_frame.
  rewrite !(RstarC Q); eauto using gimp_frame.
Qed.

Lemma gimp_eintro sigma P Q :
  interp_escr sigma = (P, Q) ->
  gimp Q (Rescr sigma).
Proof.
  red; ins; eapply GM_eintro with (r1 := res_emp); try edone; instantiate;
  rewrite ?res_plus_emp_l; repeat eexists; eauto using res_plus_emp_r; try red; ins; desf.
  by destruct Q.
Qed.

Lemma gimp_eelim sigma P Q :
  interp_escr sigma = (P, Q) ->
  gimp (Rstar P (Rescr sigma)) Q.
Proof.
  red; ins; desf; eapply GM_eelim; eauto.
    by eapply res_escr_plus; ins; eauto.
  by ins; desf; rewrite res_plusC in *; eapply RP_ext; eauto.
Qed.


(******************************************************************************)
(** ** 2.5.4. Hoare logic *)
(******************************************************************************)

(** *** Structural rules *)
(******************************************************************************)

Lemma rule_conseq :
  forall P P' e QQ QQ'
         (PRE: gimp P' P) 
         (T: triple P e QQ) 
         (POST: forall n, gimp (QQ n) (QQ' n)),
    triple P' e QQ'.
Proof.
  red; ins.
  red in T; desf; eexists P'0,; ins; eauto using gimp_trans. 
  eapply T0 in SAT'; clear PRE T0; desf; revert SAT'.
  instantiate (1 := n); clear - POST.
  induction[e r'] n; ins; desf; unfold NW; intuition; desf. 
    specialize (POST val). 
    by eauto using GM_trans. 
  exploit FORK; eauto; intro; desf; repeat eexists; eauto.

  exploit STEP; eauto; intro; desf; do 2 eexists; eauto; ins.
  exploit x1; eauto; intro; desf; eauto.
Qed.

Lemma rule_frame : 
  forall P e QQ F (T: triple P e QQ),
    triple (Rstar P F) e (fun v => Rstar (QQ v) F).
Proof.
  unfold triple; ins; desf.
  eexists (Rstar P' F),; ins; desf; eauto using gimp_frame.
  specialize (T0 _ SAT'0 n); clear - T0 SAT'1 SAT'.
  induction[e r1 T0 SAT'] n; ins; desc; unfold NW; repeat split; ins; desf.
    by eauto using GM_frame. 
  {
    exploit FORK; eauto; instantiate; clear - SAT'1 SAT' IHn; intro M; desf.
    eexists (res_plus r0 r2), r3; repeat (split; eauto). 
    rewrite res_plusC, <- res_plusA; f_equal; apply res_plusC.
    eapply IHn; eauto.
    rewrite res_plusC, <- res_plusA, (res_plusC r2) in SAT'.
    by intro X; rewrite X in SAT'.
  }

  rewrite res_plusA in RELY.
  exploit STEP; eauto; clear STEP FORK END. 
  intro X; desf.
  eexists,; eauto; ins; specialize (X0 _ IN' WPE); desf.
  rewrite <- res_plusA in X0; eexists,,; eauto.
  eapply IHn; eauto.
  by eapply guar_def in X0; intro Z; rewrite Z in *; desf.
Qed.

(** *** Pure reductions *)
(******************************************************************************)

Lemma rule_val P v : 
  triple P (Evalue v) (fun z => if z == v then P else Rfalse). 
Proof.
  eexists,; eauto using gimp_refl; ins.
  destruct n; ins; unfold NW; repeat split; ins; desf; try destruct CTX; desf;
  eauto using ghost_move. 
Qed.

Lemma rule_seq : 
  forall P e QQ (T: triple P e QQ) e' QQ'
    (T': forall v, triple (QQ v) (e' v) QQ'),
    triple P (Elet e e') QQ'. 
Proof.
  ins; red in T; desf; eexists P',; ins; clear T.
  specialize (T0 _ SAT' n); clear SAT'.
  induction[r' e T0] n; ins; desf; unfold NW; repeat (split; ins).
    by inv CTX; specialize (FORK _ CTX0 _ eq_refl); desf; eauto 7. 
  desf.
    exploit STEP; eauto; instantiate; clear - IHn; intro X; desf.
    eexists,; eauto; ins; exploit X0; eauto; ins; desf; eauto.
  specialize (END _ eq_refl); clear FORK STEP. 
  ins; desf.
  red in T'; specialize (T' n0); desf.

  eexists (Rstar P'0 (Rsingl rF)),; ins; eauto; desf.
    apply GM_frame; eauto using GM_trans; split; eauto using res_plus_emp_r.
    by destruct r'; ins; desf.
  ins; desf.
  rewrite (res_plusC rF), <- res_plusA in *.
  repeat eexists; ins; eapply T'0, RP_ext; eauto.
  by intro X; rewrite X in *.
Qed.

Lemma rule_fork P e : 
  triple P e (fun _ => Rtrue) ->
  triple P (Efork e) (fun _ => Rtrue).
Proof.
  unfold triple; ins; desf.
  eexists P',; ins.
  destruct n; ins; unfold NW; repeat split; ins; desf.
  destruct CTX; ins; desf.
  eexists res_emp,; rewrite res_plus_emp_l; repeat split; eauto.
  destruct n; ins; unfold NW; repeat split; ins; desf; ins; try destruct CTX; vauto.
Qed.

(** *** Allocation *)
(******************************************************************************)

Lemma rule_alloc_one : 
    triple (Rtrue) (Ealloc 0) (fun z => Runinit z). 
Proof.
  eexists,; eauto using gimp_refl; intros. 
  destruct n; hnf; try done; unfold NW; repeat split; try done.
    by inversion 1.
  ins; desf; ins; desf. 
  eexists,; [by apply GM_singl|]; ins; desf.
  eexists (res_mupd (res_plus r' r'1) l 0 p_uninit),; repeat split; try done.
    rewrite (res_plusC (res_plus _ _)), <- res_plusA in IN', WPE |- *.
    rewrite (res_plusC r').
    destruct (res_plus r'1 r'); ins; desf; ins; f_equal.
      by extensionality loc; inv_lplus loc; desf; ins; desf.
    apply lift_plus_none_inv in Heq; desf; desf; inv_lplus l0; desf.
    rewrite WPE in *; ins.
    by simpls; unfold prot_plus in *; desf.

  destruct n; hnf; try done; unfold NW; repeat split; ins; desf; desf;
    try by inv CTX.
  rewrite res_plusC; eapply GM_base.
  ins; rewrite res_plusC, <- res_plusA in IN'.
  destruct (res_plus r'1 r'); ins; clear IN'; desf; desf.
  exfalso; omega.
Qed.

(** *** Nonatomics *)
(******************************************************************************)

Lemma rule_read_na l v :
  triple (Rpt l v) (Eload ATna l) (fun z => Rconj (Rpt l v) (Rpure (z = v))). 
Proof.
  eexists,; eauto using gimp_refl; intros. 
  destruct n; hnf; try done; unfold NW; repeat split; try done.
    by inversion 1.
  ins; desf; ins; desf.
  eexists,; [apply GM_singl; congruence|]; ins; desf. 
  rewrite res_plusC, <- res_plusA; eexists,,; repeat split. 
  exists v; rewrite res_plusA, res_plusC; rewrite res_plusA in *.
  rewrite res_get_plus_na; eauto.
  destruct n; hnf; try done; unfold NW; repeat split; ins; desf; desf;
    try by inv CTX.
  rewrite res_plusC; eapply GM_base.
  eapply RP_ext; try split; rewrite ?Rpt_eq; ins; try split; try kundef.
    symmetry; eapply RELY1.
    by destruct r'; ins; desf; inv_lplus l; rewrite SAT' in *; ins; desf.
  rewrite res_plusC, <- res_plusA in *; kundef.
Qed.

Lemma rule_write_na l v v' :
  triple (Rpt l v) (Estore ATna l v') (fun _ => Rpt l v'). 
Proof.
  eexists,; eauto using gimp_refl; intros. 
  destruct n; hnf; try done; unfold NW; repeat split; try done.
    by inversion 1.
  ins; desf; ins; desf.
  eexists,; [apply GM_singl; congruence|]; ins; desf. 
  rewrite res_plusA in *.
  rewrite <- res_plus_upd_l; ins.
  rewrite <- res_plusA, res_plusC, <- res_plusA; eexists,,; repeat split.
    by right; exists v; rewrite res_get_plus_na; eauto.

  rewrite <- res_plusA, res_plusC, res_plusAC, <- res_plusA in IN'.
  rewrite res_plusC, res_plus_upd_l; ins; try kundef.
    destruct n; hnf; try done; unfold NW; repeat split; ins; desf; desf;
      try by inv CTX.
    by apply GM_base, res_get_upds; kundef.
  by destruct r'; ins; desf; ins; desf; inv_lplus l; rewrite SAT' in *; ins; desf.
  by destruct r'; ins; desf; ins; desf; inv_lplus l; rewrite SAT' in *; ins; desf.
Qed.

Lemma rule_write_na_uninit l v' :
  triple (Runinit l) (Estore ATna l v') (fun _ => Rpt l v'). 
Proof.
  eexists,; eauto using gimp_refl; intros. 
  destruct n; hnf; try done; unfold NW; repeat split; try done.
    by inversion 1.
  ins; desf; ins; desf.
  eexists,; [apply GM_singl; congruence|]; ins; desf. 
  rewrite res_plusA in *.
  rewrite <- res_plus_upd_l; ins.
  rewrite <- res_plusA, res_plusC, <- res_plusA; eexists,,; repeat split.
    left; destruct r'; ins; desf; ins; clear Heq Heq2.
    by inv_lplus l; rewrite SAT' in *; ins; desf.
  rewrite <- res_plusA, res_plusC, res_plusAC, <- res_plusA in IN'.
  rewrite res_plusC, res_plus_upd_l; ins; try kundef.
    destruct n; hnf; try done; unfold NW; repeat split; ins; desf; desf;
      try by inv CTX.
    by apply GM_base, res_get_upds; kundef.
  by destruct r'; ins; desf; ins; desf; inv_lplus l; rewrite SAT' in *; ins; desf.
  by destruct r'; ins; desf; ins; desf; inv_lplus l; rewrite SAT' in *; ins; desf.
Qed.

(** Acquire/release protocol rules *)
(******************************************************************************)

Lemma rule_read_at : 
  forall tau s P QQ 
         (IMP: forall s' (FUT: prot_trans tau s s') z r 
                      (SAT: Rstar (interp_prot tau s' z) P r), 
                 Rbox (QQ z) r) l,
    triple (Rstar (Rat l s tau) P) (Eload ATat l) 
           (fun z => Rstar P (Rbox (QQ z))).
Proof.
  eexists,; eauto using gimp_refl; intros. 
  destruct n; hnf; try done; unfold NW; repeat split; try done.
    by inversion 1.
  
  rewrite Rstar_at in SAT'; ins; desf; ins; desf.
  red in RELY; desf.
  eexists,; [by apply GM_singl|]. 

  ins; desf.
  generalize IN';
  rewrite ?(res_plusC _ rF), ?res_plusA, ?(res_plusAC rF), res_plusC. 
  eexists,,; repeat split; try done.

    rewrite !res_plusA in *.
    by destruct (res_get_plus r' l ((res_plus (res_strip rY) (res_plus r'1 rF))) SAT'0); 
       desf; eauto.

  destruct n; repeat split; unfold NW; intros; desf; try (by inv CTX).
  eapply GM_base; intros; apply Rstar_box; split; ins; desf.
     by eapply RP_ext; eauto; kundef.

  assert (QQ val (res_strip (res_plus rY r'))). {
    destruct (res_get_plus r' l rF SAT'0); desf.
      by intro X; rewrite X in *.
    specialize (RELY _ _ H); desf.
    rewrite GRF, H in TOK; ins; desf; desf.
    apply H0, TOK0 in SAT'1; desf.
    rewrite res_plusC, <- res_plusA in DEF. 
  
    eapply (IMP (get_state Srf)); repeat split; eauto.
      by eapply prot_trans_trans, get_state_lt; eauto. 
    by intro X; rewrite X in *.
  }
  rewrite <- res_plusA in *. 
  eapply RP_strip_ext; try (by intro X; rewrite X in *). 
  rewrite res_strip_plus, res_strip_idemp, res_plusC, <- res_strip_plus; ins; try kundef.
  by intro X; rewrite X in *; ins; destruct (QQ val).
Qed.

Lemma rule_store_at_uninit : 
  forall l v tau s,
    triple (Rstar (Runinit l) (interp_prot tau s v)) 
           (Estore ATat l v) 
           (fun _ => Rat l s tau).
Proof.
  eexists,; eauto using gimp_refl; intros. 
  destruct n; hnf; try done; unfold NW; repeat split; try done.
    by inversion 1.
  intros; simpl in STEP; desf; simpl in RELY; desf.
  eexists,.
    clear SAT'.
    eapply GM_frame with (Q := Rsingl r') (F := Rsingl rF); try eapply GM_base; try split; 
    eauto using res_plus_emp_r;
    by intro; desf; destruct r'.
  clear RELY1. 
    intros.
    rewrite RstarC in SAT'.
    rewrite Rstar_singl in IN'; try done. 
    simpl in IN'; desf.
    hnf in SAT'; desf.
    assert (G2 : res_get r2 l = p_uninit).
      by clear - SAT'1; ins; desf; ins; inv_lplus l; desf; ins; desf.
    clear SAT'1.
    assert (G1 : res_get r1 l = p_bot).
      by eapply res_get_plus_uninit2; eauto; rewrite res_plusC.
    assert (GF : res_get rF l = p_bot).
      eapply res_get_plus_uninit2; eauto; rewrite res_plusA in *.
      by intro M; rewrite M in *; destruct r1. 
    assert (G' : res_get r'1 l = p_bot).
      eapply res_get_plus_uninit2; eauto; rewrite res_plusC, res_plusA in *.
      rewrite (res_plusAC r2), <- res_plusA in IN'.
      by intro M; rewrite M in *.

  unfold guar, atomic_guar, NW, env_move; ins; desf. 
  exists (res_upd (res_plus r2 r'1) l (p_at tau (mk_slist_singl tau s))).
  exists (res_upd r1 l (p_at tau (mk_slist_singl tau s))).

  assert (FOO: res_plus (res_plus (res_plus r2 r'1) rF) r1 <> Rundef).
    intro X; destruct IN'; rewrite <- X; clear.
    rewrite !res_plusA, (res_plusAC _ r1); f_equal.
    by rewrite !(res_plusAC rF); eauto using f_equal, res_plusC.

  assert (EQ: res_plus (res_upd (res_plus r2 r'1) l (p_at tau (mk_slist_singl tau s))) rF 
          = res_upd (res_plus (res_plus r2 r'1) rF) l (p_at tau (mk_slist_singl tau s))).
    by apply res_plus_upd_l; ins; intro X; rewrite X in *.

  rewrite EQ, !res_get_upds; try (by intro M; rewrite M in *; desf).
  rewrite res_plus_upd_upd with (pi'' := p_at tau (mk_slist_singl tau s)); try done.
  2: by unfold prot_plus; rewrite mk_slist_idemp; desf.

  repeat eexists; eauto.
    by eapply RP_ext_upd; eauto; rewrite G1.
  f_equal; rewrite !res_plusA, (res_plusAC _ r1); f_equal.
  rewrite !(res_plusAC rF); eauto using f_equal, res_plusC.

  by left; split; ins; rewrite 2!res_get_plus_bot_r, res_get_plus_bot_l.

  by intro X; desf; rewrite res_get_plus_bot_r, res_get_plus_bot_l, G2 in X; ins.

  by ins; desf; rewrite res_get_plus_bot_r, res_get_plus_bot_l, G2 in TOK; ins.

  destruct n; ins; unfold NW; repeat split; ins; desf; try (by inv CTX).
  by eapply GM_base; ins; eexists,; try eapply res_get_upds; ins; eauto; intro X; rewrite X in *.
Qed.


Lemma rule_store_at : 
  forall tau s v P Q s''
         (GHP: gimp P (Rstar (interp_prot tau s'' v) Q))
         (IMP: forall s' (FUT: prot_trans tau s s') v' r
                      (SAT: Rstar (interp_prot tau s' v') P r), 
                 prot_trans tau s' s'') l,
    triple (Rstar (Rat l s tau) P) (Estore ATat l v) 
           (fun _ => Rstar (Rat l s'' tau) Q).
Proof.
  eexists,; eauto using gimp_refl; intros. 
  destruct n; hnf; try done; unfold NW; repeat split; try done.
    by inversion 1.
  rewrite Rstar_at in SAT'; ins; desf; ins; desf.

  assert (M: exists S1, res_get (res_plus r' rF) l = p_at tau S1 /\
               (forall s, In s (SL_list S) -> In s (SL_list S1))).
    destruct r', rF; ins; desf; ins.
    eapply lift_plus_some_inv with (l:=l) in Heq. 
    by rewrite SAT'0 in *; ins; desf; eauto 6 using In_mk_slist, in_or_app. 
  desf.
  specialize (RELY1 _ _ M); desc.

  assert (forall s0, In s0 (SL_list S1) -> prot_trans tau s0 s'').
    red in RELY1; desf; rewrite M in *; ins; desf; desf.
    eapply prot_trans_trans; [eby eapply get_state_lt2|]. 
    eapply IMP with (r := res_plus rY r'); [eby eapply get_state_lt2; eauto|];
    repeat eexists; eauto.
    by intro X; rewrite res_plusC, <- res_plusA, X in DEF.
  
  specialize (GHP _ SAT').
  eexists,.
    rewrite <- (res_plus_strip_idemp r'), res_plusA in RELY0 |- *. 
    apply GM_frame with (F := Rstar (Rat l (get_state S) tau) (Rsingl rF)); try split; 
    eauto using res_plus_emp_r. 
    by intro X; rewrite X, res_plusC in *.
    eexists (res_strip r'), rF,; ins; eauto using res_get_strip_at, get_state_in.
    by repeat split; eauto using res_plus_emp_r; intro X; rewrite X in *; destruct r'.

  intros. 
  rewrite !RstarA, !(RstarAC _ (Rat _ _ _)), Rstar_at in IN'; ins; desf.
  specialize (WPE _ _ _ M IN'0); desf.
  desf. 
  red in WPE0; desf.
  rewrite res_plusC in DEF.
  rewrite GRF in *; desf.

  destruct (mk_slist tau (s'' :: SL_list S0)) as [S''|] eqn: Z.
  2: exfalso; unfold mk_slist in *; desf; destruct n0; split; ins. 
  2: rewrite slist_ltD in *.
  2: desf; desf; eauto using prot_trans_refl; [left|right|];
     try eapply prot_trans_trans with (get_state S0); eauto using get_state_lt;
     try rewrite <- WPE0; eauto using get_state_in;
     try eapply prot_trans_trans with (get_state Sxxx); eauto;
     try eapply IMP with (r := res_plus r_env (res_plus r' rF)); try split; eauto 7 using RP_ext;
     eauto using get_state_in, prot_trans_trans, get_state_lt. 
  2: by destruct S0; eapply Sorted_total; eauto; ins; try apply prot_trans_ord.

  ins; desf.
  red in RELY1; desf; rewrite M in *; ins; desf; desf.
  assert (FOO:
    res_plus (res_upd (res_plus r0 r'1) l (p_at tau0 S'')) rF =
    res_upd (res_plus (res_plus r0 r'1) rF) l (p_at tau0 S'')).
    apply res_plus_upd_l2.
      by intro X; destruct IN'3; rewrite res_plusAC, res_plusC, X.
      rewrite !(res_plusAC rF) in IN'0.
    eapply prot_plus_get_helper with (r' := res_plus r1 (res_plus r0 r'1)); try done.
    rewrite IN'0.
    by eauto using prot_plus_at_bounded, in_cons, In_mk_slist.
    by rewrite IN'0.

  eexists (res_upd (res_plus r0 r'1) l (p_at tau0 S'')),
          (res_upd r1 l (p_at tau0 S'')); unfold NW; repeat split; ins.
  red; unfold NW.

{
  eexists,s'',S''; repeat (eexists; try edone); rewrite ?res_get_upds; try done.

  eapply RP_ext_upd; eauto.
   eapply prot_plus_get_helper; eauto; try done.
    by eauto using prot_plus_at_bounded, in_cons, In_mk_slist.

  erewrite FOO, res_plus_upd_upd; f_equal; eauto.
   by rewrite res_plusC, !res_plusA; do 2 f_equal; apply res_plusC.
  by rewrite res_plusC, !res_plusA, (res_plusC _ rF). 
  by ins; desf; desf; rewrite mk_slist_idemp in *; desf.

  rewrite FOO, res_get_upds; ins. 
  by rewrite res_plusA, (res_plusC _ rF); intro X; rewrite X in *; destruct r0.

  by intro; desf.

  right; eexists,; eauto; split.
    intros; rewrite slist_ltD in *.
    intros; eapply prot_trans_trans; [eby eapply get_state_lt|]. 
    intros; eapply prot_trans_trans; eauto.
    eapply IMP with (r := res_plus r_env (res_plus r' rF)); try split; 
      eauto 7 using RP_ext;
      eauto using get_state_in, prot_trans_trans, get_state_lt.

  unfold mk_slist in *; desf; ins.
  by rewrite in_undup_iff, In_sort; ins; intuition.
}

  red in RELY; desf; rewrite res_get_upds; try (by destruct r0); try (by intro; desf).
  rewrite M, GRF0 in *; ins; desf; split; ins; red; ins.
  eexists s'',; eauto using In_mk_slist, in_eq.
  eapply prot_trans_trans; [eby eapply get_state_lt|]. 
  eapply IMP with (r := res_plus rE r'); [eby eapply get_state_lt2; eauto|].
      repeat eexists; eauto.
      by intro X; rewrite res_plusC, <- res_plusA, X in DEF1.
   
  destruct n; ins; unfold NW; repeat split; ins; desf; try by inv CTX.
  eapply GM_base.
  rewrite Rstar_at; split.
  eapply RP_ext_upd; eauto using RP_ext. 
     rewrite !(res_plusAC rF), <- res_plusA, res_plusC in IN'0.
   
     eapply prot_plus_get_helper; eauto; try done.
      by eauto using prot_plus_at_bounded, in_cons, In_mk_slist.
   
  by rewrite res_get_upds; repeat eexists; eauto using In_mk_slist, in_eq.
Qed.


Lemma rule_cas : 
  forall tau s vo vn P Q R
    (GHP: forall s' (FUT: prot_trans tau s s'),
          gimp (Rstar (interp_prot tau s' vo) P)
               (Rexists (fun s'' => Rconj (Rpure (prot_trans tau s' s''))
                                (Rstar (interp_prot tau s'' vn) Q)))) l
    (IMP: forall s'' (FUT: prot_trans tau s s'') y (NEQ: y <> vo),
                 implies (Rstar (interp_prot tau s'' y) P) (Rbox R)),
  triple (Rstar (Rat l s tau) P) (Ecas l vo vn) 
         (fun z => Rexists (fun s'' => 
                      Rstar (Rat l s'' tau)
                            (if z == 0 then Rstar P (Rbox R)
                             else Q))).
Proof.
  eexists,; eauto using gimp_refl; intros. 
  destruct n; hnf; try done; unfold NW; repeat split; try done.
    by inversion 1.
  rewrite Rstar_at in SAT'; ins; desf; ins; desf.

Case "cas succeeded".
  assert (M: exists S1, res_get (res_plus r' rF) l = p_at tau S1 /\
               (forall s, In s (SL_list S) -> In s (SL_list S1))).
    destruct r', rF; ins; desf; ins.
    eapply lift_plus_some_inv with (l:=l) in Heq. 
    by rewrite SAT'0 in *; ins; desf; eauto 6 using In_mk_slist, in_or_app. 
  desf.
  specialize (RELY _ _ M); red in RELY; desf. 

  rewrite M, GRF in *; ins; desc; subst tau0.
  assert (TR: prot_trans tau s (get_state Srf)). 
    by eapply M0, TOK0 in SAT'1; desf; 
       eauto using prot_trans_trans, get_state_lt.
  rewrite res_plusC, <- res_plusA in DEF |- *.

  exploit (GHP _ TR (res_plus rY r')); clear GHP.
    by split; eauto; kundef.
  intro GM.
  eexists,.
    apply GM_frame with (F := (Rsingl rF)); try split; 
    eauto using res_plus_emp_r. 
    by intro X; rewrite X, res_plusC in *.

  intros.
  destruct (res_get_plus2 _ _ (res_plus r' rF) GRF M) as (S' & GG & CONT).
    by rewrite res_plusC.
  rewrite <- res_plusA in GG.
  specialize (WPE _ _ GG); desf.
  ins; desf; rename x into s''.

  assert (WPE1: get_state Srf = get_state S').
    apply (prot_trans_ord tau).
      by apply get_state_lt, CONT; eauto using get_state_in.
    assert (X := get_state_in S'); apply CONT in X; desf; eauto using get_state_lt.
    by apply TOK0 in X; desf; eauto using get_state_lt, prot_trans_trans. 

  destruct (mk_slist tau (s'' :: SL_list Su)) as [S''|] eqn: Z.
  2: exfalso; unfold mk_slist in *; desf; destruct n0; split; ins. 
  2: desf; desf; eauto using prot_trans_refl; [left|right|];
     try eapply prot_trans_trans with (get_state Su); eauto using get_state_lt;
     try rewrite <- WPE0, <- WPE1; eauto using get_state_in.
  2: by destruct Su; eapply Sorted_total; eauto; ins; try apply prot_trans_ord.

  clear IN'4 IN'5.
  assert (FOO:
    res_plus (res_upd (res_plus r3 r'1) l (p_at tau S'')) rF =
    res_upd (res_plus (res_plus r3 r'1) rF) l (p_at tau S'')).
    rewrite !res_plusA, (res_plusAC r3), !(res_plusC _ r'1), (res_plusAC r'1),
        <- res_plusA in IN', WPE.

    apply res_plus_upd_l2.
      by intro X; destruct IN'; rewrite res_plusAC, res_plusC, X.
    eapply prot_plus_get_helper with (r' := res_plus (res_plus r3 r'1) r0); try done.
    rewrite res_plusC, res_plusA, WPE.
    by eauto using prot_plus_at_bounded, in_cons, In_mk_slist.
    by rewrite res_plusC, res_plusA, WPE.

  eexists (res_upd (res_plus r3 r'1) l (p_at tau S'')),
          (res_upd r0 l (p_at tau S'')); unfold NW; repeat split; ins; eauto.
  2: congruence.

{
  red; unfold NW.
  rewrite FOO, !res_get_upds, res_plus_upd_upd with (pi'' := p_at tau S'');
  try (by simpl; rewrite mk_slist_idemp; desf); eauto.

  eexists,s'',S''; repeat (eexists; try edone); rewrite ?res_get_upds; try done.

    eapply RP_ext_upd; eauto.
    rewrite res_plusA in *.
    eapply prot_plus_get_helper; eauto; try done.
    by eauto using prot_plus_at_bounded, in_cons, In_mk_slist.

  by rewrite res_plusC, !res_plusA, (res_plusC rF).

  right; eexists,; eauto; split.
    intros; eapply prot_trans_trans; eauto.
    eapply prot_trans_trans; [eby eapply get_state_lt|]. 
    by rewrite <- WPE0, WPE1; apply prot_trans_refl.
  unfold mk_slist in *; desf; ins.
  by rewrite in_undup_iff, In_sort; ins; intuition.
        
  by rewrite res_plusC; revert IN'; clear; rewrite !res_plusA, (res_plusC rF).
  by intro; desf.
  intro X; destruct IN'; clear - X.
  by rewrite !res_plusA in *; rewrite (res_plusC rF), X, res_plusC.
}

  destruct n; ins; unfold NW; repeat split; ins; desf; desf; try by inv CTX.
  eapply GM_base.
  rewrite !res_plusA, (res_plusAC r3), !(res_plusC _ r'1), (res_plusAC r'1),
          <- res_plusA in IN', WPE.
  split; [by destruct (res_plus r3 r'1)|].
  eexists s''.
  rewrite Rstar_at; split.
  eapply RP_ext_upd; try apply RP_ext; eauto.
    by destruct (res_plus r3 r'1).
   
     eapply prot_plus_get_helper; eauto; try done.
      by eauto using prot_plus_at_bounded, in_cons, In_mk_slist.
   
  by rewrite res_get_upds; repeat eexists; eauto using In_mk_slist, in_eq;
     destruct (res_plus r3 r'1).

Case "cas failed".
  eexists,; [by apply GM_singl|]. 
  ins; desf.
  generalize IN';
    rewrite ?(res_plusC _ rF), ?res_plusA, ?(res_plusAC rF), res_plusC. 
  eexists,,; repeat split; try done.
  
    rewrite !res_plusA in *.
    by destruct (res_get_plus r' l ((res_plus (res_strip rY) (res_plus r'1 rF))) SAT'0); 
       desf; eauto.
  
  destruct n; repeat split; unfold NW; intros; desf; try (by inv CTX); try done.
  eapply GM_base; intros.
  split; [by intro X; rewrite X in *|].
  exists s.

  assert (R (res_strip (res_plus rY r'))). {
    destruct (res_get_plus r' l rF SAT'0); desf.
      by intro X; rewrite X in *.
    specialize (RELY _ _ H); red in RELY; desf.
    rewrite GRF, H in TOK; ins; desf; desf.
    apply H0, TOK0 in SAT'1; desf.
    rewrite res_plusC, <- res_plusA in DEF. 
  
    eapply (IMP (get_state Srf)); repeat split; eauto.
      by eapply prot_trans_trans, get_state_lt; eauto. 
    by intro X; rewrite X in *.
  }

  rewrite Rstar_at, Rstar_box; repeat split.
    by eapply RP_ext; eauto; intro X; rewrite X in *.

    rewrite <- res_plusA in *. 
    eapply RP_strip_ext; try (by intro X; rewrite X in *). 
    rewrite res_strip_plus, res_strip_idemp, res_plusC, <- res_strip_plus; ins;
      intro X; rewrite X in *; ins. 
    by destruct R. 

  destruct (res_get_plus r' l (res_plus (res_strip rY) r'1) SAT'0) 
    as (? & -> & ?); eauto; kundef.
Qed.

